from abc import ABC, abstractmethod
from types import MethodType, FunctionType
from typing import Union

from thornfield.caches.cache import Cache


class CacheFactory(ABC):
    @abstractmethod
    def create(self, func: Union[MethodType, FunctionType]) -> Cache:
        pass

    @staticmethod
    def _func_to_key(func: Union[MethodType, FunctionType]) -> str:
        return f"{func.__module__}.{func.__qualname__}"



from time import time
from typing import Optional, Any, Callable

from .cache import Cache
from ..constants import NOT_FOUND
from ..errors import CachingError
from ..postgresql_key_value_adapter import PostgresqlKeyValueAdapter

try:
    from psycopg2.pool import AbstractConnectionPool
except ImportError:
    AbstractConnectionPool = ""


class PostgresqlCache(Cache):
    def __init__(
        self,
        connection_pool: AbstractConnectionPool,
        table: str,
        serializer: Optional[Callable[[Any], str]] = None,
        deserializer: Optional[Callable[[Optional[str]], Any]] = None,
    ) -> None:
        super().__init__(serializer=serializer, deserializer=deserializer)
        self._adapter = PostgresqlKeyValueAdapter(connection_pool, table)

    def get(self, key):
        t = round(time() * 1000)
        try:
            value = self._adapter.get(self._serialize(key), t)
            if value is None:
                return NOT_FOUND
            return self._deserialize(value)
        except Exception as e:
            raise CachingError(f"Could not get {key}", exc=e)

    def set(self, key, value, expiration: int):
        if expiration:
            expiration = round(time() * 1000) + expiration
        try:
            self._adapter.set(self._serialize(key), self._serialize(value), expiration)
        except Exception as e:
            raise CachingError(f"Could not set {key} as {value}", exc=e)



from functools import partial, wraps
from inspect import getsource, iscoroutinefunction, getfullargspec
from types import MethodType
from typing import Optional, Callable, Any, Dict, List

from .caches.cache import Cache
from .constants import NOT_FOUND
from .errors import CachingError
from .typing import NotCached, Cached


class Cacher:
    def __init__(self, cache_impl: Optional[Callable[[Callable], Cache]]) -> None:
        """

        :param cache_impl: An optional factory function that gets
            the cached method and returns an implementation of ``Cache``.
        """
        super().__init__()
        self._cache_impl = cache_impl

    def cached(
        self,
        cache: Optional[Cache] = None,
        validator: Optional[Callable[[Any], bool]] = None,
        expiration: int = 0,
    ):
        """
        :param cache: The ``Cache`` to use. If ``None``, ``self.cache_impl`` is called to create one.
        :param validator: A ``callable`` that will be called on the return value of the cached function.
            The value will be cached only if ``validator`` returns ``True``.
        :param expiration: Expiration time for each key, in milliseconds.
        """
        if callable(cache):
            return self._cached(cache, None, None, 0)
        return partial(
            self._cached, cache=cache, validator=validator, expiration=expiration
        )

    def cache_method(
        self,
        method: MethodType,
        cache: Optional[Cache] = None,
        validator: Optional[Callable[[Any], bool]] = None,
        expiration: int = 0,
        use_base_method: bool = True,
    ):
        """

        :param method: The method to cache.
        :param cache: The ``Cache`` to use. If ``None``, ``self.cache_impl`` is called to create one.
        :param validator: A ``callable`` that will be called on the return value of the cached function.
            The value will be cached only if ``validator`` returns ``True``.
        :param expiration: Expiration time for each key, in milliseconds.
        :param use_base_method: If ``method`` is an implementation of a base method,
            whether to pass to ``self.cache_impl`` it or the base method.
        :return:
        """
        func_passed_to_cache = None
        if use_base_method:
            func_passed_to_cache = next(
                getattr(c, method.__name__)
                for c in reversed(method.__self__.__class__.mro())
                if hasattr(c, method.__name__)
            )
        new_method = self._cached(
            method.__func__,
            cache=cache,
            validator=validator,
            expiration=expiration,
            func_passed_to_cache=func_passed_to_cache,
        )
        setattr(
            method.__self__, method.__name__, MethodType(new_method, method.__self__)
        )

    def _cached(
        self,
        func: Callable,
        cache: Optional[Cache],
        validator: Optional[Callable[[Any], bool]],
        expiration: int,
        func_passed_to_cache: Optional[Callable] = None,
    ):
        if cache is None and self._cache_impl is None:
            raise CachingError("No cache and no cache creator provided.")

        spec = getfullargspec(func)
        func_args = spec.args
        func_annotations = spec.annotations
        if spec.defaults:
            func_defaults = dict(zip(func_args[-len(spec.defaults) :], spec.defaults))
        else:
            func_defaults = {}

        def _x(*args, **kwargs):
            is_instance_func = args and hasattr(args[0], func.__name__)
            key = self._get_key(
                is_instance_func,
                func_args,
                args,
                kwargs,
                func_defaults,
                func_annotations,
            )
            if not hasattr(func, "cache"):
                func.cache = cache or self._cache_impl(func_passed_to_cache or func)
            result = func.cache.get(key)
            if result is NOT_FOUND:
                result = func(*args, **kwargs)
                if validator is None or validator(result):
                    func.cache.set(key, result, expiration)
            return result

        source = self._get_inner_code(getsource(_x), func)
        g = dict(globals())
        g.update(locals())
        exec(compile(source, "", "exec"), g, locals())
        return wraps(func)(locals()["x"])

    @classmethod
    def _get_inner_code(cls, base, func):
        source = base[base.index("(") :]
        if iscoroutinefunction(func):
            source = f"async def x {source}"
            i = source.index("func(")
            source = f"{source[:i]}await {source[i:]}"
        else:
            source = f"def x {source}"
        return source

    @classmethod
    def _get_key(
        cls,
        is_instance_func: bool,
        func_args: List[str],
        args: tuple,
        kwargs: dict,
        defaults: Dict[str, Any],
        annotations: Dict[str, Any],
    ) -> tuple:
        key_args = func_args[1:] if is_instance_func else func_args
        if any(a is NotCached for a in annotations.values()):
            key_args = filter(lambda a: annotations.get(a) is not NotCached, key_args)
        elif any(a is Cached for a in annotations.values()):
            key_args = filter(lambda a: annotations.get(a) is Cached, key_args)
        arg_values = dict(zip(func_args, args))
        arg_values.update(kwargs)
        return tuple(
            arg_values[a] if a in arg_values else defaults[a] for a in key_args
        )



import json
from abc import ABC, abstractmethod
from time import time
from typing import Optional, Callable, Any

from .volatile_value import VolatileValue
from ..constants import NOT_FOUND
from ..errors import CachingError

try:
    from yasoo import serialize, deserialize
except ImportError:
    serialize = deserialize = None


class Cache(ABC):
    def __init__(
        self,
        serializer: Optional[Callable[[Any], str]] = None,
        deserializer: Optional[Callable[[Optional[str]], Any]] = None,
        serialization_needed: bool = True,
    ) -> None:
        super().__init__()
        if not serialization_needed:
            return
        if serialize is None or deserialize is None:
            if serializer is None or deserializer is None:
                raise CachingError(
                    'Package "yasoo" is not installed and no de/serializer passed'
                )
        self._serialize = self._default_serialize if serializer is None else serializer
        self._deserialize = (
            self._default_deserialize if deserializer is None else deserializer
        )

    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def set(self, key, value, expiration: int):
        pass

    @staticmethod
    def _to_volatile(value, expiration: int) -> VolatileValue:
        return VolatileValue(value, round(time() * 1000) + expiration)

    @staticmethod
    def _from_volatile(value: VolatileValue):
        if time() * 1000 <= value.expiration:
            return value.value
        return NOT_FOUND

    @staticmethod
    def _default_serialize(obj) -> str:
        return json.dumps(serialize(obj, preserve_iterable_types=True))

    @staticmethod
    def _default_deserialize(data: Optional[str]):
        if data is None:
            return data
        return deserialize(json.loads(data))



import re
from types import MethodType, FunctionType
from typing import Union, Optional, Callable, Any

from psycopg2.pool import AbstractConnectionPool

from .cache_factory import CacheFactory
from ..caches.postgresql_cache import PostgresqlCache
from ..postgresql_key_value_adapter import PostgresqlKeyValueAdapter


class PostgresqlCacheFactory(CacheFactory):
    def __init__(
        self,
        connection_pool: AbstractConnectionPool,
        serializer: Optional[Callable[[Any], str]] = None,
        deserializer: Optional[Callable[[Optional[str]], Any]] = None,
        index_table: str = "_index",
    ) -> None:
        super().__init__()
        self.connection_pool = connection_pool
        self.serializer = serializer
        self.deserializer = deserializer
        self._adapter = PostgresqlKeyValueAdapter(
            self.connection_pool, index_table, ts_col=None
        )

    def create(self, func: Union[MethodType, FunctionType]) -> PostgresqlCache:
        table = self._normalize_table_name(self._func_to_key(func))
        return PostgresqlCache(
            connection_pool=self.connection_pool,
            table=table,
            serializer=self.serializer,
            deserializer=self.deserializer,
        )

    def _normalize_table_name(self, table_name: str) -> str:
        if len(table_name) > 63:
            n = self._adapter.get(table_name)
            if not n:
                used = {int(self._adapter.get(k)) for k in self._adapter.keys()}
                n = min({i + 1 for i in range(len(used) + 1)}.difference(used))
                self._adapter.set(table_name, str(n))
            return f"cache_{n}"
        else:
            return re.sub(r"[^\w]", "_", table_name)



from typing import Callable, Optional, List

try:
    from psycopg2.pool import AbstractConnectionPool
except ImportError:
    AbstractConnectionPool = ""


class ConnectionWrapper:
    def __init__(
        self, connection, release_callback: Callable[["ConnectionWrapper"], None]
    ) -> None:
        super().__init__()
        self._connection = connection
        self._callback = release_callback

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._callback(self)

    def execute_query(self, query: str, fetch: int, *params: str):
        with self._connection.cursor() as cursor:
            cursor.execute(query, params)
            if fetch == 1:
                return cursor.fetchone()
            elif fetch == -1:
                return cursor.fetchall()
        self._connection.commit()


class ConnectionPoolWrapper:
    def __init__(self, pool: AbstractConnectionPool) -> None:
        super().__init__()
        self._pool = pool

    def getconn(self) -> ConnectionWrapper:
        return ConnectionWrapper(self._pool.getconn(), self.putconn)

    def putconn(self, conn: ConnectionWrapper):
        self._pool.putconn(conn)


class PostgresqlKeyValueAdapter:
    def __init__(
        self,
        connection_pool: AbstractConnectionPool,
        table_name: str,
        key_col: str = "key",
        value_col: str = "value",
        ts_col: Optional[str] = "ts",
    ) -> None:
        super().__init__()
        self._pool = ConnectionPoolWrapper(connection_pool)
        self._table = table_name
        self._key_col = key_col
        self._value_col = value_col
        self._ts_col = ts_col

        self._create_table_if_not_exists()

    def set(self, key: str, value: str, ts: Optional[int] = None):
        if self._ts_col:
            assert ts is not None
        with self._pool.getconn() as connection:
            exists = connection.execute_query(
                f"select exists(select * from {self._table} where {self._key_col}=%s)",
                1,
                key,
            )
            if exists[0]:
                self._update(connection, key, value, ts)
            else:
                self._add(connection, key, value, ts)

    def get(self, key: str, max_ts: Optional[int] = None) -> Optional[str]:
        query = f"select {self._value_col} from {self._table} where {self._key_col}=%s"
        if max_ts is not None:
            assert self._ts_col
            query += f" and {self._ts_col}<{max_ts}"
        with self._pool.getconn() as connection:
            result = connection.execute_query(query, 1, key)
        return result[0] if result else None

    def keys(self) -> List[str]:
        with self._pool.getconn() as connection:
            result = connection.execute_query(
                f"select {self._key_col} from {self._table}", -1
            )
        return [t[0] for t in result]

    def _create_table_if_not_exists(self):
        with self._pool.getconn() as connection:
            exists = connection.execute_query(
                f"select exists(select * from information_schema.tables where table_name=%s)",
                1,
                self._table,
            )
            if not exists[0]:
                structure = f"id serial primary key, {self._key_col} text unique, {self._value_col} text"
                if self._ts_col:
                    structure += f", {self._ts_col} bigint"
                connection.execute_query(f"create table {self._table} ({structure})", 0)

    def _update(self, connection: ConnectionWrapper, key, value, ts):
        values_str = f"{self._value_col}=%s"
        if self._ts_col:
            values_str += f", {self._ts_col}={ts}"
        connection.execute_query(
            f"update {self._table} set ({values_str}) where {self._key_col}=%s",
            0,
            value,
            key,
        )

    def _add(self, connection: ConnectionWrapper, key, value, ts):
        columns = f"{self._key_col}, {self._value_col}"
        values = "%s, %s"
        if self._ts_col:
            columns += f", {self._ts_col}"
            values += f", {ts}"
        connection.execute_query(
            f"insert into {self._table} ({columns}) values ({values})", 1, key, value,
        )
